// Copyright (C) 2013 S&P Capital IQ
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package com.spcapitaliq.platformexecuting

import java.io.File
import java.net.URI

import scalaz.{\/, NonEmptyList}
import scalaz.std.string._
import scalaz.syntax.apply._
import scalaz.syntax.equal._
import scalaz.syntax.id._
import scalaz.syntax.std.option._
import scalaz.syntax.semigroup._
import NonEmptyList.{nel, nels}

import Platform._
import Util.map2

/** Finding installed and jar-contained executables. */
object Capture {
  /** Directories that might contain a binary. */
  private def execPath: Seq[String] =
    (Option(System getenv "PATH").toSeq
     flatMap (_ split File.pathSeparator filterNot (_.isEmpty)))

  private[platformexecuting] lazy val isWindows =
    System getProperty "os.name" startsWith "Windows"

  /** Alternate names for ordinary binary executable files on this
    * platform.
    */
  private def thisPlatformExes(base: String): NonEmptyList[String] =
    if (isWindows) nels(base |+| ".exe")
    else nels(base)

  /** Alternate names for ordinary binary executable files on `plat`.
    */
  private def platformExes(base: String, plat: Platform): NonEmptyList[String] =
    if (plat.os === "windows") nels(base |+| ".exe")
    else nels(base)

  /** If an executable location is explicitly specified in the
    * environment, sanity check it.  None means unspecified, Some Left
    * means insane.
    */
  def explicitExe(env: String): Option[String \/ File] =
    Option(System getenv env) map (ef =>
      (Some(new File(ef)) filter (_.exists)) \/>
        ("%s: not found; unset or fix %s first"
           format (ef, env)))

  /** Search the exec path for `exe`. */
  def implicitExe(exe: String, env: Option[String] = None
                ): String \/ File =
    (map2(execPath.view, thisPlatformExes(exe).list.view)(new File(_, _))
       find (_.exists)) \/> ("%s: not found; add directory to PATH%s"
                               format (exe, env map (" or set " |+| _) orZero))

  /** If `env` is set in the environment, prefer it, otherwise
    * searching PATH.  Answer a reason it couldn't be found, or a path
    * to the `exe`.
    */
  def localExe(exe: String, env: Option[String]): String \/ File =
    env flatMap explicitExe getOrElse implicitExe(exe, env)

  /** Where an executable ought to be stored in the classpath.
    */
  def captureTo(exe: String, resourceRoot: String, platform: Platform): String =
    capturePaths(exe, resourceRoot, platform).head

  /** Classloader resource paths at which an executable might be
    * found.
    */
  def capturePaths(exe: String, resourceRoot: String, platform: Platform
                  ): NonEmptyList[String] =
    ^(nel(platform, compatPlatforms(platform).toList),
      platformExes(exe, platform)){(osarch, pe) =>
      "%s/%s/%s/%s" format (resourceRoot, osarch.os, osarch.arch, pe)
    }

  /** If `exe` has been captured, and is appropriate for this
    * platform, yield its location.
    */
  def capturedExe(exe: String, resourceRoot: String): Option[URI] = {
    val ccl = Thread.currentThread.getContextClassLoader
    platformSpec flatMap (ps =>
      capturePaths(exe, resourceRoot, ps).list.view
        map (pth => Option(ccl getResource pth))
        collectFirst {case Some(rs) => rs.toURI})
  }
}
